#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "qheader.h"

#define WIDTH_10THS 85
#define HEIGHT_10THS 110

#define private static

#ifndef __linux__
typedef unsigned char __u_char;
typedef unsigned short __u_short;
typedef unsigned int __u_int;
typedef unsigned long __u_long;
#endif

/*
 * Default X and Y resolution.
 */
#define X_DPI 72
#define Y_DPI 72

#if BYTE_ORDER == BIG_ENDIAN
#define assign_ushort(a,v) a = ((v) >> 8) + ((v) << 8)
#else
#define assign_ushort(a,v) a = v
#endif

typedef struct gif_header_s
{
    __u_char signature[3];		/* magic number == 'GIF' */
    __u_char version[3];		/* version # '87a' or '89a' */
    __u_short width;		/* screen width */
    __u_short height;		/* screen height */

#define globalcolor_shift 7
#define colorres_shift 4
#define sort_shift 3
#define colorsize_shift 0

    __u_char flags;
    __u_char background;		/* background color index */
    __u_char aspect;		/* pixel aspect ratio */
} gif_header;

typedef struct image_descriptor_s
{
    __u_short left_pos;		/* image left pos (pixels) */
    __u_short top_pos;		/* image top  pos (pixels) */
    __u_short width;		/* image width    (pixels) */
    __u_short height;		/* image height   (pixels) */
    __u_char flags;
}
image_descriptor;

/********************************************************/
/* LZW routines are based on:                           */
/* Dr. Dobbs Journal --- Oct. 1989.                     */
/* Article on LZW Data Compression by Mark R. Nelson    */
/********************************************************/

#define MAX_BITS 12		/* this is max for GIF. */

#define TABLE_SIZE 5123		/* this is max for 12-bit codes */
#define TABLE_HASH_SHIFT 2	/* size < 4095 + (4095 >> shift) */

/* State of LZW encoder */
typedef struct code_entry_s
{
    int code_value;
    __u_short prefix_code;
    __u_char append_character;
}
code_entry;
typedef struct lzw_encoder_s
{
    int bits;
    __u_short Max_Code;
    __u_short Clear_code;
    __u_short next_code;
    FILE *file;
    code_entry *table;
    __u_short string_code;
    /* State of output buffer */
    __u_char output_bit_buffer;
    int output_bit_count;	/* # of valid low-order bits */
    /* (between 0 and 7) in buffer */
    __u_int byte_count;
    __u_char gif_buffer[260];
}
lzw_encoder;

/* Initialize LZW encoder */
private void lzw_set_bits (register lzw_encoder *, int);
private void lzw_reset (register lzw_encoder *);
private int
 lzw_init (register lzw_encoder * pe, int bits, FILE * file)
{
    lzw_set_bits (pe, bits);
    pe->Clear_code = (1 << bits);
    pe->file = file;
    pe->byte_count = 1;
    pe->output_bit_count = 0;
    pe->output_bit_buffer = 0;
    pe->table = (code_entry *) malloc (TABLE_SIZE * sizeof (code_entry));

    if (pe->table == 0)
	return -2;

    lzw_reset (pe);
    pe->string_code = 0;
    return 0;
}
/* Establish the width of the code in bits */
private void
 lzw_set_bits (register lzw_encoder * pe, int bits)
{
    pe->bits = bits;
    pe->Max_Code = (1 << (bits + 1)) - 1;
}
/* Reset the encoding table */
private void
 lzw_reset (register lzw_encoder * pe)
{
    int index;

    for (index = 0; index < TABLE_SIZE; index++)
	pe->table[index].code_value = -1;
    pe->next_code = pe->Clear_code + 2;
}

/* Put out (data) of length (bits) to GIF buffer */
private void
 lzw_putc (register lzw_encoder * pe, __u_int data)
{
    int bits = pe->bits + 1;	/* output width */
    __u_long buffer = pe->output_bit_buffer | ((__u_long) data << pe->output_bit_count);

    pe->output_bit_count += bits;
    while (pe->output_bit_count >= 8)
    {				/* putc(output_bit_buffer >> 24, file); */
	pe->gif_buffer[pe->byte_count] = (__u_char) buffer;		/* low byte */
	buffer >>= 8;
	pe->output_bit_count -= 8;
	pe->byte_count++;
	if (pe->byte_count == 256)
	{
	    pe->byte_count = 1;
	    pe->gif_buffer[0] = 255;	/* byte count for block */
	    fwrite (pe->gif_buffer, 1, 256, pe->file);
	}
    }
    pe->output_bit_buffer = (__u_char) buffer;
}

/* Finish encoding, and flush the buffers. */
private void
 lzw_finish (register lzw_encoder * pe)
{
    lzw_putc (pe, pe->string_code);	/* output last code */
    lzw_putc (pe, pe->Clear_code + 1);	/* output eof code */
    lzw_putc (pe, 0);		/* force out last code */
    if (pe->byte_count != 1)
    {
	pe->gif_buffer[0] = pe->byte_count;
	fwrite (pe->gif_buffer, 1, pe->byte_count + 1, pe->file);
    }
}

/* Terminate LZW encoder. */
private void
 lzw_exit (register lzw_encoder * pe)
{
    free ((char *) pe->table);
}

/* Get the next (depth) bits from the pixel buffer. */
/* Note that 8 % depth == 0. */
/* Free variables: bits_left, bit_buffer, next, depth, depth_mask. */
#define lzw_getc()\
 (bits_left =\
  (bits_left == 0 ?\
   (bit_buffer = *(next++), 8 - (depth)) :\
   bits_left - (depth)),\
  (bit_buffer >> bits_left) & (depth_mask))

/* Output 1 row of data in GIF (LZW) format. */
private void
 lzw (__u_char * from, __u_char * end, register lzw_encoder * pe, int depth)
{
    int bits_left = 0;
    __u_int bit_buffer = 0;
    __u_char *next = from;
    __u_int depth_mask = (1 << depth) - 1;

    if (pe->next_code == (pe->Clear_code + 2))	/* first time through */
    {
	pe->string_code = lzw_getc ();
    }

    while (next < end || bits_left >= depth)
    {
	__u_int data = lzw_getc ();	/* actually only a byte */

	/* Hash to find a match for the prefix+char */
	/* string in the string table */

	__u_short hash_prefix = pe->string_code;
	int index = (data << 4) ^ hash_prefix;
	int hash_offset;
	register code_entry *pce;

	index += index >> TABLE_HASH_SHIFT;
	if (index == 0)
	    hash_offset = 1;
	else
	    hash_offset = TABLE_SIZE - index;

	while (1)
	{
	    pce = &pe->table[index];
	    if (pce->code_value == -1)
		break;
	    if (pce->prefix_code == hash_prefix &&
		pce->append_character == data)
		break;
	    index -= hash_offset;
	    if (index < 0)
		index += TABLE_SIZE;
	}
	if (pce->code_value != -1)
	    pe->string_code = pce->code_value;
	else
	{			/* Make a new entry */
	    pce->code_value = pe->next_code++;
	    pce->prefix_code = pe->string_code;
	    pce->append_character = data;

	    lzw_putc (pe, pe->string_code);

	    if (pe->next_code > (pe->Max_Code + 1))
	    {			/* Increment the width of the code */
		if (pe->bits + 1 >= MAX_BITS)
		{		/* output clear code first */
		    lzw_putc (pe, pe->Clear_code);
		    pe->bits = (depth == 1 ? 2 : depth);
		    lzw_reset (pe);
		}
		else
		    pe->bits++;
		lzw_set_bits (pe, pe->bits);
	    }
	    pe->string_code = data;
	}
    }
}

/* Write a page to a file in GIF format. */
int DoGif (FILE * file, __u_char * faximg, __u_char *row, short raster, 
	   short ysize, short npage)
{
    static unsigned char monopal[6] =
    {0xff, 0xff, 0xff, 0, 0, 0};
    __u_short height = ysize;
    int depth = 1;
    __u_short gif_width = raster * (8 / depth);	/* decoders want the width */

    /* on a byte boundary */
    __u_char *end = row + raster;
    gif_header header;
    image_descriptor header_desc;
    lzw_encoder encoder;
    int y;
    int code = 0;		/* return code */

    /* Put the message on stderr so it doesn't interfere with */
    /* possible piped output. */

    if (row == 0)		/* can't allocate row buffer */
	return -2;
    code = lzw_init (&encoder, (depth == 1 ? 2 : depth), file);
    if (code < 0)
	return code;

    /* Set up the header. */

    memcpy (header.signature, "GIF", 3);
    memcpy (header.version, "87a", 3);
    assign_ushort (header.width, gif_width);
    assign_ushort (header.height, height);
/*      header.flags.globalcolor = TRUE;        */
/*      header.flags.colorres = depth-1;        */
/*      header.flags.sort = FALSE;              */
/*      header.flags.colorsize = depth-1;       */
    header.flags =
	(1 << globalcolor_shift) +
	((depth - 1) << colorres_shift) +
	(0 << sort_shift) +
	((depth - 1) << colorsize_shift);
    header.background = 0;
    header.aspect = 0;

    /* Write the header, on the first page only. */

    if (npage == 0)
    {
	fprintf (stderr, "The Graphics Interchange Format(c) is\n");
	fprintf (stderr,
		 "the Copyright Property of CompuServe Incorporated.\n");
	fprintf (stderr, 
	    "GIF(sm) is a Service Mark property of CompuServe Incorporated.\n");

	if (fwrite (&header, 1, 13, file) < 13)
	{
	    code = -2;
	    goto gif_done;
	}

	/* Write the header global color palette. */
	fwrite (monopal, 1, 6, file);
    }

    header_desc.left_pos = 0;
    header_desc.top_pos = 0;
    assign_ushort (header_desc.width, gif_width);
    assign_ushort (header_desc.height, height);
/*      header_desc.flags.localcolor = TRUE;    */
/*      header_desc.flags.interlace = FALSE;    */
/*      header_desc.flags.sort = FALSE;         */
/*      header_desc.flags.localsize = depth - 1; */

    header_desc.flags =
	(1 << globalcolor_shift) +
	((depth - 1) << colorsize_shift);

    /* Write the header image descriptor. */

    fputc (0x2c, file);		/* start with separator */
    if (fwrite (&header_desc, 1, 9, file) < 9)
    {
	code = -2;
	goto gif_done;
    }

    /* Write the local color palette. */
    fwrite (monopal, 1, 6, file);

    fputc (encoder.bits, file);	/* start with code size */

    lzw_putc (&encoder, encoder.Clear_code);	/* output clear code first */

    /* Dump the contents of the image. */
    for (y = 0; y < height; y++)
    {
	memcpy (row, faximg, 216);
	faximg += 216;
	lzw (row, end, &encoder, depth);
    }

    lzw_finish (&encoder);
    fputc (0, file);

gif_done:
    lzw_exit (&encoder);
    return code;
}
void exitgif (FILE * file)
{
    if (file != NULL)
    {
	fputc (0x3b, file);	/* EOF indicator */
    }
}
